# -*- coding: utf-8 -*-
# *** FORMAT PARAMS ***
import functools
from typing import overload, List, Any  # noqa
from nptyping import NDArray, Float  # noqa

import numpy as np  # noqa

from .base import Py5Base
from .image import Py5Image  # noqa
from jpype.types import JException, JArray, JBoolean, JInt, JFloat  # noqa
from .pmath import _numpy_to_pvector, _numpy_to_pmatrix2d, _numpy_to_pmatrix3d  # noqa


def _return_py5shader(f):
    @functools.wraps(f)
    def decorated(self_, *args):
        return Py5Shader(f(self_, *args))
    return decorated


def _load_py5shader(f):
    @functools.wraps(f)
    def decorated(self_, *args):
        try:
            return Py5Shader(f(self_, *args))
        except JException as e:
            msg = e.message()
            if msg == 'None':
                msg = 'shader file cannot be found'
        raise RuntimeError('cannot load shader file ' +
                           str(args[0]) + '. error message: ' + msg)
    return decorated


def _py5shader_set_wrapper(f):
    @functools.wraps(f)
    def decorated(self_, name, *args):
        args = list(args)
        if isinstance(args[0], np.ndarray):
            array = args[0]
            if array.shape in [(2,), (3,)]:
                args[0] = _numpy_to_pvector(array)
            elif array.shape == (2, 3):
                args[0] = _numpy_to_pmatrix2d(array)
            elif array.shape == (4, 4):
                args[0] = _numpy_to_pmatrix3d(array)
        else:
            def fix_type(arg):
                if isinstance(arg, bool):
                    return JBoolean(arg)
                elif isinstance(arg, int):
                    return JInt(arg)
                elif isinstance(arg, float):
                    return JFloat(arg)
                else:
                    return arg
            args = [fix_type(a) for a in args]
        return f(self_, name, *tuple(args))
    return decorated


class Py5Shader(Py5Base):
    """This class encapsulates a GLSL shader program, including a vertex and a fragment
    shader.

    Underlying Java class: PShader.PShader

    Notes
    -----

    This class encapsulates a GLSL shader program, including a vertex and a fragment
    shader. It's compatible with the P2D and P3D renderers, but not with the default
    renderer. Use the ``load_shader()`` function to load your shader code. [Note:
    It's strongly encouraged to use ``load_shader()`` to create a PShader object,
    rather than calling the PShader constructor manually.]
"""

    def __init__(self, pshader):
        self._instance = pshader
        super().__init__(instance=pshader)

    @overload
    def set(self, name: str, x: bool, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, x: bool, y: bool, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, x: bool, y: bool, z: bool, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, x: bool, y: bool, z: bool, w: bool, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, vec: JArray(JBoolean), /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, boolvec: JArray(
            JBoolean), ncoords: int, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, x: float, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, x: float, y: float, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, x: float, y: float, z: float, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, x: float, y: float,
            z: float, w: float, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, vec: NDArray[(Any,), Float], /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, vec: NDArray[(
            Any,), Float], ncoords: int, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, x: int, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, x: int, y: int, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, x: int, y: int, z: int, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, x: int, y: int, z: int, w: int, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, vec: JArray(JInt), /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, vec: JArray(JInt), ncoords: int, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, tex: Py5Image, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, mat: NDArray[(2, 3), Float], /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, mat: NDArray[(4, 4), Float], /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @overload
    def set(self, name: str, mat: NDArray[(
            4, 4), Float], use3x3: bool, /) -> None:
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        pass

    @_py5shader_set_wrapper
    def set(self, *args):
        """Sets the uniform variables inside the shader to modify the effect while the
        program is running.

        Underlying Java method: PShader.set

        Methods
        -------

        You can use any of the following signatures:

         * set(name: str, boolvec: JArray(JBoolean), ncoords: int) -> None
         * set(name: str, mat: NDArray[(2, 3), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float]) -> None
         * set(name: str, mat: NDArray[(4, 4), Float], use3x3: bool) -> None
         * set(name: str, tex: Py5Image) -> None
         * set(name: str, vec: JArray(JBoolean)) -> None
         * set(name: str, vec: JArray(JInt)) -> None
         * set(name: str, vec: JArray(JInt), ncoords: int) -> None
         * set(name: str, vec: NDArray[(Any,), Float]) -> None
         * set(name: str, vec: NDArray[(Any,), Float], ncoords: int) -> None
         * set(name: str, x: bool) -> None
         * set(name: str, x: bool, y: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool) -> None
         * set(name: str, x: bool, y: bool, z: bool, w: bool) -> None
         * set(name: str, x: float) -> None
         * set(name: str, x: float, y: float) -> None
         * set(name: str, x: float, y: float, z: float) -> None
         * set(name: str, x: float, y: float, z: float, w: float) -> None
         * set(name: str, x: int) -> None
         * set(name: str, x: int, y: int) -> None
         * set(name: str, x: int, y: int, z: int) -> None
         * set(name: str, x: int, y: int, z: int, w: int) -> None

        Parameters
        ----------

        boolvec: JArray(JBoolean)
            missing variable description

        mat: NDArray[(2, 3), Float]
            matrix of values

        mat: NDArray[(4, 4), Float]
            matrix of values

        name: str
            the name of the uniform variable to modify

        ncoords: int
            number of coordinates per element, max 4

        tex: Py5Image
            sets the sampler uniform variable to read from this image texture

        use3x3: bool
            enforces the matrix is 3 x 3

        vec: JArray(JBoolean)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: JArray(JInt)
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        vec: NDArray[(Any,), Float]
            modifies all the components of an array/vector uniform variable. PVector can only be used if the type of the variable is vec3.

        w: bool
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: float
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        w: int
            fourth component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[4], vec4)

        x: bool
            first component of the variable to modify

        x: float
            first component of the variable to modify

        x: int
            first component of the variable to modify

        y: bool
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: float
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        y: int
            second component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[2], vec2)

        z: bool
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: float
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        z: int
            third component of the variable to modify. The variable has to be declared with an array/vector type in the shader (i.e.: int[3], vec3)

        Notes
        -----

        Sets the uniform variables inside the shader to modify the effect while the
        program is running.
"""
        return self._instance.set(*args)



import sys
import os
from multiprocessing import Process
from pathlib import Path
import re
import tempfile
import textwrap

from . import jvm
from . import reference as ref


_CODE_FRAMEWORK_EXTRAS = """
def _py5_init_dynamic_variables(sketch):
    sketch.load_pixels()
    sketch.load_np_pixels()
{0}

def _py5_update_dynamic_variables(sketch):
{0}

py5._py5sketch._add_pre_hook('setup', '_py5_init_dynamic_variables', _py5_init_dynamic_variables)
py5._py5sketch._add_pre_hook('draw', '_py5_update_dynamic_variables', _py5_update_dynamic_variables)
"""


_CODE_FRAMEWORK = """
import py5

{2}

with open('{0}', 'r') as f:
    eval(compile(f.read(), '{0}', 'exec'))

py5.run_sketch(block=True)
if {1} and py5.is_dead_from_error:
    py5.exit_sketch()
"""


_STANDARD_CODE_TEMPLATE = """
import py5

def settings():
    py5.size({0}, {1}, py5.{2})


def setup():
{4}

    py5.save_frame("{3}")
    py5.exit_sketch()
"""


_ALT_CODE_TEMPLATE = """
import py5

def settings():
    py5.size({0}, {1}, py5.{2}, "{3}")


def setup():
{4}

    py5.exit_sketch()
"""


_DXF_CODE_TEMPLATE = """
import py5

def settings():
    py5.size({0}, {1}, py5.P3D)


def setup():
    py5.begin_raw(py5.DXF, "{3}")

{4}

    py5.end_raw()
    py5.exit_sketch()
"""

SETTINGS_REGEX = re.compile(r'^def settings\(\):', flags=re.MULTILINE)
SETUP_REGEX = re.compile(r'^def setup\(\):', flags=re.MULTILINE)
SETUP_CODE_REGEX = re.compile(
    r'^def setup\(\):.*?(?=^\w|\Z)',
    flags=re.MULTILINE | re.DOTALL)
DRAW_REGEX = re.compile(r'^def draw\(\):', flags=re.MULTILINE)
CODE_REGEXES = {
    f: re.compile(
        r'^\s*(' + f + r'\([^\)]*\))',
        flags=re.MULTILINE) for f in [
            'size',
            'full_screen',
            'smooth',
            'no_smooth',
        'pixel_density']}


def fix_triple_quote_str(code):
    for m in re.finditer(r'\"\"\"[^\"]*\"\"\"', code):
        code = code.replace(
            m.group(), m.group().replace('\n    ', '\n'))
    return code


def prepare_code(code):
    "transform functionless or setttings-less py5 code into code that runs"
    if SETTINGS_REGEX.search(code):
        return False, code
    no_setup = SETUP_REGEX.search(code) is None
    no_draw = DRAW_REGEX.search(code) is None

    # get just the setup function if it is defined
    code2 = code if no_setup else SETUP_CODE_REGEX.search(code).group()
    # find the key lines in the relevant code
    matches = [m for m in [r.search(code2)
                           for r in CODE_REGEXES.values()] if m]

    # if anything was found, build the settings function
    if matches:
        lines = [(m.start(), m.group(1)) for m in matches]
        settings = 'def settings():\n'
        for start, line in sorted(lines):
            settings += f'    {line}\n'
            # replace the original line so it doesn't get called in setup
            code = code.replace(line, f'pass  # moved to settings(): {line}')
    else:
        settings = ''

    if no_setup and no_draw:
        # put all of the remaining code into a setup function
        remaining_code = 'def setup():\n' + textwrap.indent(code, prefix='    ')
        remaining_code = fix_triple_quote_str(remaining_code)
    else:
        # remaining code has been modified with key lines moved from setup to
        # settings
        remaining_code = code

    return True, f'{settings.strip()}\n\n{remaining_code.strip()}\n'


def run_sketch(
        sketch_path,
        classpath=None,
        new_process=False,
        exit_if_error=False):
    sketch_path = Path(sketch_path)
    if not sketch_path.exists():
        print(f'file {sketch_path} not found')
        return

    with open(sketch_path, 'r') as f:
        code = f.read()
    tranformed, code = prepare_code(code)
    if tranformed:
        temp_py = sketch_path.with_suffix('.tmp.py')
        with open(temp_py, 'w') as f:
            f.write(code)
        sketch_path = temp_py

    def _run_sketch(sketch_path, classpath, exit_if_error):
        if not jvm.is_jvm_running():
            if classpath:
                jvm.add_classpath(classpath)
            jvm.add_jars(sketch_path.parent / 'jars')

        import py5
        if not py5.get_current_sketch().is_ready:
            py5.reset_py5()
        sys.path.extend([str(sketch_path.absolute().parent), os.getcwd()])
        py5_ns = dict()
        py5_ns.update(py5.__dict__)
        update_dynamic_variables_code = '\n'.join(
            f'    global {v}\n    {v} = sketch.{v}' for v in ref.UPDATE_DYNAMIC_VARIABLES)
        exec(
            _CODE_FRAMEWORK.format(
                sketch_path,
                exit_if_error,
                _CODE_FRAMEWORK_EXTRAS.format(update_dynamic_variables_code)),
            py5_ns)

    if new_process:
        p = Process(
            target=_run_sketch,
            args=(
                sketch_path,
                classpath,
                exit_if_error))
        p.start()
        return p
    else:
        _run_sketch(sketch_path, classpath, exit_if_error)
        if tranformed:
            os.remove(temp_py)


def run_single_frame_sketch(renderer, code, width, height, user_ns, safe_exec):
    if renderer == 'SVG':
        template = _ALT_CODE_TEMPLATE
        suffix = '.svg'
        read_mode = 'r'
    elif renderer == 'PDF':
        template = _ALT_CODE_TEMPLATE
        suffix = '.pdf'
        read_mode = 'rb'
    elif renderer == 'DXF':
        template = _DXF_CODE_TEMPLATE
        suffix = '.dxf'
        read_mode = 'r'
    else:
        template = _STANDARD_CODE_TEMPLATE
        suffix = '.png'
        read_mode = 'rb'

    import py5
    if not py5.get_current_sketch().is_ready:
        py5.reset_py5()

    if safe_exec:
        prepared_code = textwrap.indent(code, '    ')
        prepared_code = fix_triple_quote_str(prepared_code)
    else:
        user_ns['_py5_user_ns'] = user_ns
        code = code.replace('"""', r'\"\"\"')
        prepared_code = f'    exec("""{code}""", _py5_user_ns)'

    temp_py = tempfile.NamedTemporaryFile(suffix='.py')
    temp_out = tempfile.NamedTemporaryFile(suffix=suffix)

    with open(temp_py.name, 'w') as f:
        code = template.format(
            width,
            height,
            renderer,
            temp_out.name,
            prepared_code)
        f.write(code)

    exec(_CODE_FRAMEWORK.format(temp_py.name, True, ''), user_ns)

    if not safe_exec:
        del user_ns['_py5_user_ns']

    py5.reset_py5()

    temp_py.close()

    with open(temp_out.name, read_mode) as f:
        result = f.read()

    temp_out.close()
    return result


__all__ = ['run_sketch', 'run_single_frame_sketch']



import sys
import time
import threading
from collections.abc import Iterable
from typing import Callable, Dict, Tuple, Any

from .. import methods


class Py5Promise:

    def __init__(self):
        self._result = None
        self._is_ready = False

    @property
    def is_ready(self) -> bool:
        return self._is_ready

    @property
    def result(self) -> Any:
        return self._result

    def _set_result(self, result):
        self._result = result
        self._is_ready = True


class Py5Thread:

    def __init__(self, sketch, f, args, kwargs):
        self.sketch = sketch
        self.f = f
        self.args = args
        self.kwargs = kwargs

    def stop(self):
        pass

    def __call__(self):
        try:
            self.f(*self.args, **self.kwargs)
        except Exception:
            methods.handle_exception(*sys.exc_info())
            self.sketch._terminate_sketch()


class Py5PromiseThread(Py5Thread):

    def __init__(self, sketch, f, promise, args, kwargs):
        super().__init__(sketch, f, args, kwargs)
        self.promise = promise

    def stop(self):
        super().stop()

    def __call__(self):
        try:
            self.promise._set_result(self.f(*self.args, **self.kwargs))
        except Exception:
            methods.handle_exception(*sys.exc_info())
            self.sketch._terminate_sketch()


class Py5RepeatingThread(Py5Thread):

    def __init__(self, sketch, f, delay, args, kwargs):
        super().__init__(sketch, f, args, kwargs)
        self.repeat = True
        self.delay = delay
        self.e = threading.Event()

    def stop(self):
        super().stop()
        self.repeat = False
        self.e.set()

    def __call__(self):
        try:
            while self.repeat:
                start_time = time.time()
                self.f(*self.args, **self.kwargs)
                self.e.wait(max(0, start_time + self.delay - time.time()))
        except Exception:
            self.stop()
            methods.handle_exception(*sys.exc_info())
            self.sketch._terminate_sketch()


class ThreadsMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._py5threads = {}

    def _check_param_types(self, args, kwargs):
        if not isinstance(args, Iterable) and args is not None:
            raise RuntimeError(
                'args argument must be iterable (such as a tuple or list)')
        if not isinstance(kwargs, dict) and kwargs is not None:
            raise RuntimeError('kwargs argument must be a dictionary')

        kwargs = kwargs or {}
        args = args or ()

        return args, kwargs

    def _launch_py5thread(self, name, py5thread, daemon):
        if isinstance(py5thread, Py5RepeatingThread) and self.has_thread(name):
            self.stop_thread(name, wait=True)

        t = threading.Thread(name=name, target=py5thread, daemon=daemon)
        t.start()
        self._py5threads[t.name] = (t, py5thread)

        return t.name

    def _shutdown(self):
        self.stop_all_threads(wait=False)
        super()._shutdown()

    # *** BEGIN METHODS ***

    def launch_thread(self, f: Callable, name: str = None, daemon: bool = True,
                      args: Tuple = None, kwargs: Dict = None) -> str:
        """new template no description.

        Parameters
        ----------

        args: Tuple
            missing variable description

        daemon: bool
            missing variable description

        f: Callable
            missing variable description

        kwargs: Dict
            missing variable description

        name: str
            missing variable description

        Notes
        -----

        new template no description.
"""
        args, kwargs = self._check_param_types(args, kwargs)
        return self._launch_py5thread(
            name, Py5Thread(
                self, f, args, kwargs), daemon)

    def launch_promise_thread(
            self,
            f: Callable,
            name: str = None,
            daemon: bool = True,
            args: Tuple = None,
            kwargs: Dict = None) -> Py5Promise:
        """new template no description.

        Parameters
        ----------

        args: Tuple
            missing variable description

        daemon: bool
            missing variable description

        f: Callable
            missing variable description

        kwargs: Dict
            missing variable description

        name: str
            missing variable description

        Notes
        -----

        new template no description.
"""
        args, kwargs = self._check_param_types(args, kwargs)
        promise = Py5Promise()
        self._launch_py5thread(
            name, Py5PromiseThread(
                self, f, promise, args, kwargs), daemon)
        return promise

    def launch_repeating_thread(
            self,
            f: Callable,
            name: str = None,
            time_delay: float = 0,
            daemon: bool = True,
            args: Tuple = None,
            kwargs: Dict = None) -> str:
        """new template no description.

        Parameters
        ----------

        args: Tuple
            missing variable description

        daemon: bool
            missing variable description

        f: Callable
            missing variable description

        kwargs: Dict
            missing variable description

        name: str
            missing variable description

        time_delay: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        args, kwargs = self._check_param_types(args, kwargs)
        return self._launch_py5thread(
            name, Py5RepeatingThread(
                self, f, time_delay, args, kwargs), daemon)

    def _remove_dead_threads(self):
        thread_names = list(self._py5threads.keys())
        for t_name in thread_names:
            if not self._py5threads[t_name][0].is_alive():
                del self._py5threads[t_name]

    def has_thread(self, name: str) -> None:
        """new template no description.

        Parameters
        ----------

        name: str
            missing variable description

        Notes
        -----

        new template no description.
"""
        self._remove_dead_threads()
        return name in self._py5threads

    def stop_thread(self, name: str, wait: bool = False) -> None:
        """new template no description.

        Parameters
        ----------

        name: str
            missing variable description

        wait: bool
            missing variable description

        Notes
        -----

        new template no description.
"""
        if name in self._py5threads:
            t, py5thread = self._py5threads[name]
            py5thread.stop()
            if wait:
                t.join()
            del self._py5threads[name]

    def stop_all_threads(self, wait: bool = False) -> None:
        """new template no description.

        Parameters
        ----------

        wait: bool
            missing variable description

        Notes
        -----

        new template no description.
"""
        current_thread_name = threading.current_thread().name
        for name in self.list_threads():
            if name == current_thread_name:
                # don't try to join a thread with itself
                continue
            self.stop_thread(name, wait=wait)

    def list_threads(self) -> None:
        """new template no description.

        Notes
        -----

        new template no description.
"""
        self._remove_dead_threads()
        return list(self._py5threads.keys())



from pathlib import Path
from typing import overload, List, Union  # noqa

import numpy as np
from PIL import Image
import jpype


class PixelMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._instance = kwargs['instance']
        self._np_pixels = None

    def _replace_instance(self, new_instance):
        self._instance = new_instance
        super()._replace_instance(new_instance)

    def _init_np_pixels(self):
        width = self.pixel_width
        height = self.pixel_height
        self._py_bb = bytearray(width * height * 4)
        self._java_bb = jpype.nio.convertToDirectBuffer(self._py_bb)
        self._np_pixels = np.asarray(
            self._py_bb, dtype=np.uint8).reshape(
            height, width, 4)

    # *** BEGIN METHODS ***

    def load_np_pixels(self) -> None:
        """new template no description.

        Notes
        -----

        new template no description.
"""
        if self._np_pixels is None:
            self._init_np_pixels()
        self._instance.loadPixels()
        self._java_bb.asIntBuffer().put(self._instance.pixels)

    def update_np_pixels(self) -> None:
        """new template no description.

        Notes
        -----

        new template no description.
"""
        if self._np_pixels is None:
            self._init_np_pixels()
        self._java_bb.asIntBuffer().get(self._instance.pixels)
        self._instance.updatePixels()

    @property
    def np_pixels(self) -> np.ndarray:
        """new template no description.

        Notes
        -----

        new template no description.
"""
        return self._np_pixels

    def set_np_pixels(self, array: np.ndarray, bands: str = 'ARGB') -> None:
        """new template no description.

        Parameters
        ----------

        array: np.ndarray
            missing variable description

        bands: str
            missing variable description

        Notes
        -----

        new template no description.
"""
        self.load_np_pixels()
        if bands == 'L':
            self._np_pixels[:, :, 0] = 255
            self._np_pixels[:, :, 1:] = array[:, :,
                                              None] if array.ndim == 2 else array
        elif bands == 'ARGB':
            self._np_pixels[:] = array
        elif bands == 'RGB':
            self._np_pixels[:, :, 0] = 255
            self._np_pixels[:, :, 1:] = array
        elif bands == 'RGBA':
            self._np_pixels[:, :, 0] = array[:, :, 3]
            self._np_pixels[:, :, 1:] = array[:, :, :3]
        self.update_np_pixels()

    def save(self,
             filename: Union[str,
                             Path],
             format: str = None,
             drop_alpha: bool = True,
             **params) -> None:
        """new template no description.

        Parameters
        ----------

        drop_alpha: bool
            missing variable description

        filename: Union[str, Path]
            missing variable description

        format: str
            missing variable description

        params
            missing variable description

        Notes
        -----

        new template no description.
"""
        filename = Path(str(self._instance.savePath(str(filename))))
        self.load_np_pixels()
        arr = self.np_pixels[:, :, 1:] if drop_alpha else np.roll(
            self.np_pixels, -1, axis=2)
        Image.fromarray(arr).save(filename, format=format, **params)



import time
import io
from pathlib import Path
import tempfile

from IPython.display import display, SVG, Image
from IPython.core.magic import Magics, magics_class, cell_magic, line_magic
from IPython.core.magic_arguments import magic_arguments, argument, parse_argstring

import PIL

from .run import run_single_frame_sketch


def wait(wait_time, sketch):
    end_time = time.time() + wait_time
    while time.time() < end_time and sketch.is_running:
        time.sleep(0.2)


class BaseHook:

    def __init__(self, hook_name):
        self.hook_name = hook_name
        self.is_ready = False
        self.exception = None
        self.is_terminated = False

    def hook_finished(self, sketch):
        sketch._remove_post_hook('draw', self.hook_name)
        self.is_ready = True

    def hook_error(self, sketch, e):
        self.exception = e
        sketch._remove_post_hook('draw', self.hook_name)
        self.is_terminated = True

    def sketch_terminated(self):
        self.is_terminated = True


class ScreenshotHook(BaseHook):

    def __init__(self, filename):
        super().__init__('py5screenshot_hook')
        self.filename = filename

    def __call__(self, sketch):
        try:
            sketch.save_frame(self.filename)
            self.hook_finished(sketch)
        except Exception as e:
            self.hook_error(sketch, e)


class SaveFramesHook(BaseHook):

    def __init__(self, dirname, filename, start, limit):
        super().__init__('py5save_frames_hook')
        self.dirname = dirname
        self.filename = filename
        self.start = start
        self.limit = limit
        self.num_offset = None
        self.filenames = []

    def __call__(self, sketch):
        try:
            if self.num_offset is None:
                self.num_offset = 0 if self.start is None else sketch.frame_count - self.start
            num = sketch.frame_count - self.num_offset
            frame_filename = sketch._insert_frame(
                str(self.dirname / self.filename), num=num)
            sketch.save_frame(frame_filename)
            self.filenames.append(frame_filename)
            if len(self.filenames) == self.limit:
                self.hook_finished(sketch)
        except Exception as e:
            self.hook_error(sketch, e)


class GrabFramesHook(BaseHook):

    def __init__(self, delay, count):
        super().__init__('py5grab_frames_hook')
        self.delay = delay
        self.count = count
        self.frames = []
        self.last_frame_time = 0

    def __call__(self, sketch):
        try:
            if time.time() - self.last_frame_time < self.delay / 1000:
                return
            sketch.load_np_pixels()
            self.frames.append(sketch.np_pixels[:, :, 1:].copy())
            self.last_frame_time = time.time()
            if len(self.frames) == self.count:
                self.hook_finished(sketch)
        except Exception as e:
            self.hook_error(sketch, e)


@magics_class
class Py5Magics(Magics):

    def _filename_check(self, filename):
        filename = Path(filename)
        if not filename.parent.exists():
            filename.parent.mkdir(parents=True)
        return filename

    @magic_arguments()
    @argument('width', type=int, help='width of PDF output')
    @argument('height', type=int, help='height of PDF output')
    @argument('filename', type=str, help='filename for PDF output')
    @argument('--unsafe', dest='unsafe', action='store_true',
              help="allow new variables to enter the global namespace, creating a potentially unsafe situation")
    @cell_magic
    def py5drawpdf(self, line, cell):
        """Create a PDF with py5.

        For users who are familiar with Processing and py5 programming, you can
        pretend the code in this cell will be executed in a sketch with no
        `draw()` function and your code in the `setup()` function. It will use
        the PDF renderer.

        The below example will create a red square on a gray background:

        ```
            %%py5drawpdf 500 250 /tmp/test.pdf
            py5.background(128)
            py5.fill(255, 0, 0)
            py5.rect(80, 100, 50, 50)
        ```

        As this is creating a PDF, you cannot do operations on the
        `pixels` or `np_pixels` arrays. Use `%%py5draw` instead.

        Code used in this cell can reference functions and variables defined in
        other cells. By default, variables and functions created in this cell
        will be local to only this cell because to do otherwise would be unsafe.
        If you understand the risks, you can use the `global` keyword to add a
        single function or variable to the notebook namespace or the --unsafe
        argument to add everything to the notebook namespace. Either option may
        be very useful to you, but be aware that using py5 objects in a
        different notebook cell or reusing them in another sketch can result in
        nasty errors and bizzare consequences. Any and all problems resulting
        from using these features are solely your responsibility and not the py5
        library maintainers.
        """
        args = parse_argstring(self.py5drawpdf, line)
        pdf = run_single_frame_sketch('PDF', cell, args.width, args.height,
                                      self.shell.user_ns, not args.unsafe)
        if pdf:
            filename = self._filename_check(args.filename)
            with open(filename, 'wb') as f:
                f.write(pdf)
            print(f'PDF written to {filename}')

    @magic_arguments()
    @argument('width', type=int, help='width of DXF output')
    @argument('height', type=int, help='height of DXF output')
    @argument('filename', type=str, help='filename for DXF output')
    @argument('--unsafe', dest='unsafe', action='store_true',
              help="allow new variables to enter the global namespace, creating a potentially unsafe situation")
    @cell_magic
    def py5drawdxf(self, line, cell):
        """Create a DXF file with py5.

        For users who are familiar with Processing and py5 programming, you can
        pretend the code in this cell will be executed in a sketch with no
        `draw()` function and your code in the `setup()` function. It will use
        the DXF renderer.

        The below example will create a rotated cube:

        ```
           %%py5drawdxf 200 200 /tmp/test.dxf
            py5.translate(py5.width / 2, py5.height / 2)
            py5.rotate_x(0.4)
            py5.rotate_y(0.8)
            py5.box(80)
        ```

        As this is creating a DXF file, your code will be limited to the
        capabilities of that renderer.

        Code used in this cell can reference functions and variables defined in
        other cells. By default, variables and functions created in this cell
        will be local to only this cell because to do otherwise would be unsafe.
        If you understand the risks, you can use the `global` keyword to add a
        single function or variable to the notebook namespace or the --unsafe
        argument to add everything to the notebook namespace. Either option may
        be very useful to you, but be aware that using py5 objects in a
        different notebook cell or reusing them in another sketch can result in
        nasty errors and bizzare consequences. Any and all problems resulting
        from using these features are solely your responsibility and not the py5
        library maintainers.
        """
        args = parse_argstring(self.py5drawdxf, line)
        dxf = run_single_frame_sketch('DXF', cell, args.width, args.height,
                                      self.shell.user_ns, not args.unsafe)
        if dxf:
            filename = self._filename_check(args.filename)
            with open(filename, 'w') as f:
                f.write(dxf)
            print(f'DXF written to {filename}')

    @magic_arguments()
    @argument('width', type=int, help='width of SVG drawing')
    @argument('height', type=int, help='height of SVG drawing')
    @argument('-f', '--filename', type=str, dest='filename',
              help='save SVG drawing to file')
    @argument('--unsafe', dest='unsafe', action='store_true',
              help="allow new variables to enter the global namespace, creating a potentially unsafe situation")
    @cell_magic
    def py5drawsvg(self, line, cell):
        """Create a SVG drawing with py5 and embed result in the notebook.

        For users who are familiar with Processing and py5 programming, you can
        pretend the code in this cell will be executed in a sketch with no
        `draw()` function and your code in the `setup()` function. It will use
        the SVG renderer.

        The below example will create a red square on a gray background:

        ```
            %%py5drawsvg 500 250
            py5.background(128)
            py5.fill(255, 0, 0)
            py5.rect(80, 100, 50, 50)
        ```

        As this is creating a SVG drawing, you cannot do operations on the
        `pixels` or `np_pixels` arrays. Use `%%py5draw` instead.

        Code used in this cell can reference functions and variables defined in
        other cells. By default, variables and functions created in this cell
        will be local to only this cell because to do otherwise would be unsafe.
        If you understand the risks, you can use the `global` keyword to add a
        single function or variable to the notebook namespace or the --unsafe
        argument to add everything to the notebook namespace. Either option may
        be very useful to you, but be aware that using py5 objects in a
        different notebook cell or reusing them in another sketch can result in
        nasty errors and bizzare consequences. Any and all problems resulting
        from using these features are solely your responsibility and not the py5
        library maintainers.
        """
        args = parse_argstring(self.py5drawsvg, line)
        svg = run_single_frame_sketch('SVG', cell, args.width, args.height,
                                      self.shell.user_ns, not args.unsafe)
        if svg:
            if args.filename:
                filename = self._filename_check(args.filename)
                with open(filename, 'w') as f:
                    f.write(svg)
                print(f'SVG drawing written to {filename}')
            display(SVG(svg))

    @magic_arguments()
    @argument('width', type=int, help='width of PNG image')
    @argument('height', type=int, help='height of PNG image')
    @argument('-f', '--filename', dest='filename', help='save image to file')
    @argument('-r', '--renderer', type=str, dest='renderer', default='HIDDEN',
              help='processing renderer to use for sketch')
    @argument('--unsafe', dest='unsafe', action='store_true',
              help="allow new variables to enter the global namespace, creating a potentially unsafe situation")
    @cell_magic
    def py5draw(self, line, cell):
        """Create a PNG image with py5 and embed result in the notebook.

        For users who are familiar with Processing and py5 programming, you can
        pretend the code in this cell will be executed in a sketch with no
        `draw()` function and your code in the `setup()` function. By default it
        will use the default Processing renderer.

        The below example will create a red square on a gray background:

        ```
            %%py5draw 500 250
            py5.background(128)
            py5.fill(255, 0, 0)
            py5.rect(80, 100, 50, 50)
        ```

        Code used in this cell can reference functions and variables defined in
        other cells. By default, variables and functions created in this cell
        will be local to only this cell because to do otherwise would be unsafe.
        If you understand the risks, you can use the `global` keyword to add a
        single function or variable to the notebook namespace or the --unsafe
        argument to add everything to the notebook namespace. Either option may
        be very useful to you, but be aware that using py5 objects in a
        different notebook cell or reusing them in another sketch can result in
        nasty errors and bizzare consequences. Any and all problems resulting
        from using these features are solely your responsibility and not the py5
        library maintainers.
        """
        args = parse_argstring(self.py5draw, line)

        if args.renderer == 'SVG':
            print('please use %%py5drawsvg for SVG drawings.')
            return
        if args.renderer == 'PDF':
            print('please use %%py5drawpdf for PDFs.')
            return
        if args.renderer not in ['HIDDEN', 'JAVA2D', 'P2D', 'P3D']:
            print(f'unknown renderer {args.renderer}')
            return

        png = run_single_frame_sketch(
            args.renderer,
            cell,
            args.width,
            args.height,
            self.shell.user_ns,
            not args.unsafe)
        if png:
            if args.filename:
                filename = self._filename_check(args.filename)
                PIL.Image.open(
                    io.BytesIO(png)).convert(
                    mode='RGB').save(filename)
                print(f'PNG file written to {filename}')
            display(Image(png))

    @line_magic
    @magic_arguments()
    @argument('-w', type=int, dest='wait', default=0,
              help='wait time in seconds before taking screenshot')
    def py5screenshot(self, line):
        """Take a screenshot of the current running sketch.

        Use the -w argument to wait before taking the screenshot.

        The returned image is a `PIL.Image` object. It can be assigned to a
        variable or embedded in the notebook.

        Below is an example demonstrating how to take a screenshot after a two
        second delay and assign it to the `img` variable. The image is then
        saved to a file. When run from a notebook, the image is embedded in the
        output.

        ```
            img = %py5screenshot -w 2
            img.save('image.png')
            img
        ```
        """
        args = parse_argstring(self.py5screenshot, line)
        import py5
        sketch = py5.get_current_sketch()

        if not sketch.is_running:
            print('The current sketch is not running.')
            return

        wait(args.wait, sketch)

        with tempfile.NamedTemporaryFile(suffix='.png') as png_file:
            hook = ScreenshotHook(png_file.name)
            sketch._add_post_hook('draw', hook.hook_name, hook)

            while not hook.is_ready and not hook.is_terminated:
                time.sleep(0.005)

            if hook.is_ready:
                return PIL.Image.open(png_file.name)
            elif hook.is_terminated and hook.exception:
                print('error running magic:', hook.exception)

    @line_magic
    @magic_arguments()
    @argument('dirname', type=str, help='directory to save the frames')
    @argument('--filename', type=str, dest='filename',
              default='frame_####.png', help='filename to save frames to')
    @argument('-w', type=int, dest='wait', default=0,
              help='wait time in seconds before starting sketch frame capture')
    @argument('-s', dest='start', type=int,
              help='frame starting number instead of sketch frame_count')
    @argument('--limit', type=int, dest='limit', default=0,
              help='limit the number of frames to save (default 0 means no limit)')
    def py5screencapture(self, line):
        """Save the current running sketch's frames to a directory.

        Use the -w argument to wait before starting.

        The below example will save the next 50 frames to the `/tmp/frames`
        directory after a 3 second delay. The filenames will be saved with the
        default name 'frame_####.png' with numbering that starts at 0.

        ```
            %py5screencapture /tmp/frames -w 3 -s 0 --limit 50
        ```

        If a limit is given, this line magic will wait to return a list of the
        filenames. Otherwise, it will return right away as the frames are saved
        in the background. It will keep doing so as long as the sketch continues
        to run.
        """
        args = parse_argstring(self.py5screencapture, line)
        import py5
        sketch = py5.get_current_sketch()

        if not sketch.is_running:
            print('The current sketch is not running.')
            return

        dirname = Path(args.dirname)
        if not dirname.exists():
            dirname.mkdir(parents=True)
        print(f'writing frames to {str(args.dirname)}...')

        wait(args.wait, sketch)

        hook = SaveFramesHook(dirname, args.filename, args.start, args.limit)
        sketch._add_post_hook('draw', hook.hook_name, hook)

        if args.limit:
            while not hook.is_ready and not hook.is_terminated:
                time.sleep(0.02)
                print(
                    f'saving frame {len(hook.filenames)}/{args.limit}',
                    end='\r')
            print('')

            if hook.is_ready:
                return hook.filenames

        if hook.is_terminated and hook.exception:
            print('error running magic:', hook.exception)

    @line_magic
    @magic_arguments()
    @argument('filename', type=str, help='filename of gif to create')
    @argument('count', type=int, help='number of Sketch snapshots to create')
    @argument('delay', type=int,
              help='time in milliseconds between Sketch snapshots')
    @argument('duration', type=int,
              help='time in milliseconds between frames in the GIF')
    @argument('-w', type=int, dest='wait', default=0,
              help='wait time in seconds before starting sketch frame capture')
    @argument('-l', dest='loop', type=int, default=0,
              help='number of times for the GIF to loop (default of 0 loops indefinitely')
    @argument('--optimize', action='store_true', help='optimize GIF palette')
    def py5animatedgif(self, line):
        """Save the current running sketch's frames to a directory.

        Use the -w argument to wait before starting.

        The below example will create a 10 frame animated GIF saved to
        '/tmp/animated.gif'. The frames will be recorded 1000 milliseconds
        apart after waiting 3 seconds. The animated GIF will display the frames
        with a 500 millisecond delay between each one and will loop indefinitely.

        ```
            %py5screencapture /tmp/animated.gif 10 1000 500 -w 3
        ```
        """
        args = parse_argstring(self.py5animatedgif, line)
        import py5
        sketch = py5.get_current_sketch()

        if not sketch.is_running:
            print('The current sketch is not running.')
            return

        filename = Path(args.filename)

        wait(args.wait, sketch)

        hook = GrabFramesHook(args.delay, args.count)
        sketch._add_post_hook('draw', hook.hook_name, hook)

        while not hook.is_ready and not hook.is_terminated:
            time.sleep(0.05)
            print(
                f'collecting frame {len(hook.frames)}/{args.count}',
                end='\r')
        print('')

        if hook.is_ready:
            if not filename.parent.exists():
                filename.parent.mkdir(parents=True)

            img1 = PIL.Image.fromarray(hook.frames[0], mode='RGB')
            imgs = [PIL.Image.fromarray(arr, mode='RGB')
                    for arr in hook.frames[1:]]
            img1.save(
                filename,
                save_all=True,
                duration=args.duration,
                loop=args.loop,
                optimize=args.optimize,
                append_images=imgs)

            return str(filename)

        elif hook.is_terminated and hook.exception:
            print('error running magic:', hook.exception)


def load_ipython_extension(ipython):
    ipython.register_magics(Py5Magics)
